/*****************************************************************************/
/*
                                  HTTPd.c

WASD VMS Hypertext Transfer Protocol daemon.


OBLIGATORY COPYRIGHT NOTICE
---------------------------
WASD VMS Hypertext Services, Copyright (C) 1996,1997 Mark G.Daniel.
(prior to 01-JUL-1997, "HFRD VMS Hypertext Services")

This package is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; version 2 of the License, or any later version. 

This package is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
details. 

You should have received a copy of the GNU General Public License along with 
this package; if not, write to the Free Software Foundation, Inc., 675 Mass 
Ave, Cambridge, MA 02139, USA. 


QUALIFIERS  (implemented in the CLI.c module)
----------
/ACCEPT=                comma separated list of accepted hosts/domains
/CGI_PREFIX=            prefix to CGI-script variable (symbol) names
/DBUG                   turn on all "if (Debug)" statements (if compiled DBUG)
/DO=                    command to be passed to server
/FILBUF=                number of bytes in record buffer
/FORMAT=                log record format
/REJECT=                comma separated list of rejected hosts/domains
/[NO]LOG[=]             logging enabled/disabled, optional log file name
/[NO]MONITOR            monitor logicals enabled/disabled
/NETBUF=                number of bytes in network read buffer
/OUTBUF=                number of bytes in output buffer
/PERIOD=                log file period
/PORT=                  server IP port number
/PRIORITY=              process priority
/[NO]PROFILE            allow/disallow SYSUAF-authenticated access control
/[NO]PROMISCUOUS        authenticates any user name for path authorization
/SERVICES=              list of [host-name:]port providing HTTP service
/SOFTWAREID=            overrides the server's software ID string
/SUBBUF=                number of bytes in subprocess's SYS$OUTPUT buffer
/[NO]SYSUAF             allow/disallow SYSUAF authentication
/[NO]SWAP               allow/disallow process to be swapped out
/VERSION                simply display the server version
/[NO]ZERO               accounting zeroed on startup (default is non-zeroed)


VERSION HISTORY
---------------
06-DEC-97  MGD  v4.5.1, resolved a suspected inconsistent AST delivery
                        situation by requiring all $QIO()s with an AST routine
                        to ensure any queueing errors, etc. are reported via
                        the AST routine, by an explicit $DCLAST() ... this
                        removes some ambiguity about how $QIO() returns should
                        be handled ... drastic, but desperate times, etc.
                        (a more consistent and desirable model anyway :^)
02-NOV-97  MGD  v4.5.0, file cache,
                        logging periods,
                        HttpdSupervisor(),
                        configurable script run-time environments,
                        additional request header fields
18-OCT-97  MGD  v4.4.1, bugfixes; duration, logging period
01-OCT-97  MGD  v4.4.0, message module,
                        conditional rule mapping,
                        SYSUAF-authenticated user access control,
                        multi-homed/multi-port services
                        (some NETLIB supported packages now cannot DNS lookup),
                        echo and Xray internal scripts,
                        extensions to logging functionality,
                        additional command-line server control,
                        bugfix; redirection loop detection
01-AUG-97  MGD  v4.3.0, MadGoat NETLIB broadens TCP/IP package support,
                        server activity report
16-JUL-97  MGD  v4.2.2, bugfix; WORLD realm and access list
07-JUL-97  MGD  v4.2.1, minimum heap allocation chunk size,
                        prevent keep-alive timeout redefining request logical
01-JUL-97  MGD  v4.2.0, change name to WASD (Wide Area Surveillance Division)
                        persistant DCL subprocesses and CGIplus 
                        (see re-written DCL.C module),
                        scripting and client reports,
                        potential multi-thread problems in reports fixed
27-MAR-97  MGD  v4.1.0, rationalized HTTP response header generation;
                        delete on close for "temporary" files, to support
                        UPD module "preview" functionality ... WARNING, any
                        file with a name component comprising a leading hyphen,
                        sixteen digits and a trailing hyphen will be deleted!
01-FEB-97  MGD  v4.0.0, HTTPd version 4
01-OCT-96  MGD  v3.4.0, extended server reporting  
01-AUG-96  MGD  v3.3.0, realm/path-based authorization;
                        BASIC and DIGEST authentication;
                        PUT(/POST/DELETE) module;
                        StmLf module (variable to stream-LF file conversion)
12-APR-96  MGD  v3.2.0, file record/binary now determined by record format;
                        persistent connections ("Keep-Alive" within HTTP/1.0);
                        moved RMS parse structures into thread data;
                        improved local redirection detection;
                        observed Multinet disconnection/zero-byte behaviour
                        (request now aborts if network read returns zero bytes)
15-FEB-96  MGD  v3.1.1, fixed rediculous :^( bug in 302 HTTP header;
                        minor changes to request accounting and server report;
                        minor changes for user directory support;
                        minor changes to error reporting
03-JAN-96  MGD  v3.1.0, support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-95  MGD  v3.0.0, single heap for each thread's dynamic memory management;
                        extensive rework of DCL subprocess functionality;
                        HTML pre-processsing module (aka Server Side Includes);
                        NCSA/CERN compliant image-mapping module;
                        NetWriteBuffered() for improving network IO;
                        miscellaneous reworks/rewrites
27-SEP-95  MGD  v2.3.0, carriage-control on non-header records from <CR><LF> 
                        to single <LF> ('\n' ... newline); some browsers expect
                        only this (e.g. Netscape 1.n was spitting on X-bitmaps)
                        added Greenwich Mean Time time-stamp functionality;
                        added 'Referer:', 'If-Modified-Since:', 'User-Agent:'
07-AUG-95  MGD  v2.2.2, optionally include commented VMS file specifications
                        in HTML documents and VMS-style directory listings
16-JUN-95  MGD  v2.2.1, added file type description to "Index of" (directory)
24-MAY-95  MGD  v2.2.0, minor changes to allow compilation on AXP platform
03-APR-95  MGD  v2.1.0, add SYSUAF authentication, POST method handling
20-DEC-94  MGD  v2.0.0, multi-threaded version
20-JUN-94  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#define HttpdVersion "4.5.1"

/* if not NETLIB then UCX! */
#ifndef IP_NETLIB
#  define IP_UCX 1
#  define HttpdIp "Digital-TCPIP"
#else
#  define HttpdIp "MadGoat-NETLIB"
#endif

#ifdef __ALPHA
#  define HttpdArch "AXP"
#else
#  define HttpdArch "VAX"
#endif

/* standard C header files */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <prvdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application-related header files */
#include "wasd.h"
#include "cli.h"
#include "copyright.h"
#include "error.h"
#include "graph.h"
#include "httpd.h"
#include "logging.h"
#include "mapurl.h"
#include "msg.h"
#include "net.h"
#include "request.h"
#include "support.h"

/******************/
/* global storage */
/******************/

/*
   These are the required privileges of the executing HTTP server.
   The server ACCOUNT should only have TMPMBX and NETMBX (just for
   extra security, policy ... keep privileged accounts to a minimum).
   The image should be installed with SYSPRV, PRMMBX, PSWAPM and SYSNAM.
   Subprocesses are spawned only with the process's current privileges,
   which are always maintained at TMPMBX and NETMBX.  If additional
   privileges are required for any particular purpose (e.g. binding to
   a privileged IP port) then they are enabled, the action performed,
   and then they are disabled again immediately.
*/
unsigned long  AvJoePrivMask [2] = { PRV$M_NETMBX | PRV$M_TMPMBX, 0 },
               ResetPrivMask [2] = { 0xffffffff, 0xffffffff };

char  Utility [] = "HTTPD";

/*
   Decided to be able to eliminate the debug statements from production
   executables completely.  This will eliminate some largely unused code
   from the images reducing the overall file size, but more importantly
   will eliminate the test and branch execution overhead each time a debug
   statement is encountered.  Do this by conditionally turning the integer
   Debug storage into a constant false value ... compiler optimization of
   an impossible-to-execute section of code does the rest!  Very little
   other code needs to be explicitly conditionally compiled.
*/
#ifdef DBUG
boolean  Debug;
#else
#define Debug 0
#endif

boolean  MonitorEnabled = true,
         NaturalLanguageEnglish,
         NoSwapOut = true;

int  DclSysOutputSize = DefaultDclSysOutputSize,
     ExitStatus,
     FileBufferSize = DefaultFileBufferSize,
     NetReadBufferSize = DefaultNetReadBufferSize,
     OutputBufferSize = DefaultOutputBufferSize,
     ProcessPriority = 4,
     RejectHostsLength,
     ServerPort = 80;

unsigned long  CurrentConnectCount,
               MaxConcurrentConnections = DefaultMaxConcurrentConnections;

char  BuildInfo [64],
      CommandLine [256],
      HttpdUserName [13],
      NaturalLanguage [64],
      ServerPortString [16],
      SoftwareID [256];

char  *AcceptHostsPtr = NULL,
      *RejectHostsPtr = NULL;

struct  GenericListStruct  RequestList;
struct  GenericListStruct  RequestHistoryList;
int  RequestHistoryCount,
     RequestHistoryMax;

struct AnExitHandler  ExitHandler;

struct AccountingStruct  Accounting;
boolean  AccountingZeroOnStartup;

#ifdef DBUG
   USEC_TIMER_STORAGE
#endif

/********************/
/* external storage */
/********************/

extern boolean  AuthorizationEnabled;
extern boolean  CliLoggingDisabled;
extern boolean  CliLoggingEnabled;
extern boolean  CliMonitorDisabled;
extern boolean  CliMonitorEnabled;
extern boolean  CliVersion;
extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern boolean  LoggingEnabled;
extern int  CliServerPort;
extern int  ServerPort;
extern char  *CliAcceptHostsPtr;
extern char  *CliRejectHostsPtr;
extern char  CliLogFileName[];
extern char  ControlBuffer[];
extern char  ErrorSanityCheck[];
extern char  HttpProtocol[];
extern struct ConfigStruct  Config;
extern struct MsgDbStruct  MsgDb;

#ifdef IP_UCX
   extern unsigned short  ClientChannel;
   extern unsigned short  ServerChannel;
#endif

#ifdef IP_NETLIB
   extern unsigned long  ClientNetLibSocket;
   extern unsigned long  ServerNetLibSocket;
#endif

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned long  AltpriMask [2] = { PRV$M_ALTPRI, 0 },
                         PswapmMask [2] = { PRV$M_PSWAPM, 0 };

   static struct {
      unsigned short  BufferLength;
      unsigned short  ItemCode;
      unsigned long  BufferAddress;
      unsigned long  ReturnLengthAddress;
   } JpiItem [] =
       { { sizeof(HttpdUserName)-1, JPI$_USERNAME, HttpdUserName, 0 },
         { 0, 0, 0, 0 } };

   register char  *cptr;

   int  status;
   unsigned short  Length;
   char  TimeScratch [32],
         ProcessName [16];
   $DESCRIPTOR (NaturalLanguageDsc, NaturalLanguage);
   $DESCRIPTOR (ProcessNameDsc, ProcessName);
   $DESCRIPTOR (ProcessNameFaoDsc, "HTTPd:!UL");

   /*********/
   /* begin */
   /*********/

   if (!SoftwareID[0])
      sprintf (SoftwareID, "HTTPd-WASD/%s %OpenVMS/%s %s",
               HttpdVersion, HttpdArch, HttpdIp);

#ifdef __DECC
   sprintf (BuildInfo, "%s %5.5s VMS %s DECC %d",
            __DATE__, __TIME__, __VMS_VERSION+1, __DECC_VER);
#else
   sprintf (BuildInfo, "%s %5.5s VMS %s VAXC ?",
            __DATE__, __TIME__, VMS_VERSION+1);
#endif

   if (VMSnok (status = ParseCommandLine ()))
      exit (status);

#ifdef DBUG
   if (!Debug)
      if (getenv ("HTTPD$DBUG") != NULL)
         Debug = true;
#endif

   if (ControlBuffer[0])
   {
      /*********************************************************/
      /* this image is being used to control the HTTPd process */
      /*********************************************************/

      if (CliServerPort) ServerPort = CliServerPort;
      if (ServerPort < 1 || ServerPort > 65535)
         ErrorExitVmsStatus (NO_EXIT_MSG, "IP port", FI_LI);
      exit (ControlCommand (ControlBuffer));
   }

   /****************/
   /* HTTPd server */
   /****************/

   /* output the GNU GENERAL PUBLIC LICENSE message */
   fprintf (stdout, "%%%s-I-SOFTWAREID, %s\n%s",
            Utility, SoftwareID, CopyRightMessageBrief);

   if (CliVersion) exit (SS$_NORMAL);

   if (VMSnok (status = sys$getjpi (0, 0, 0, &JpiItem, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpi()", FI_LI);
   HttpdUserName[13] = '\0';
   for (cptr = HttpdUserName; *cptr && !ISLWS(*cptr); cptr++);
   *cptr = '\0';
   if (Debug) fprintf (stdout, "HttpdUserName |%s|\n", HttpdUserName);

   /* make sure the process' privileges are those of a mere mortal */
   status = sys$setprv (0, &ResetPrivMask, 0, 0);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   status = sys$setprv (1, &AvJoePrivMask, 0, 0);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   /* set process priority */
   if (VMSnok (status = sys$setprv (1, &AltpriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   if (VMSnok (status = sys$setpri (0, 0, ProcessPriority, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setpri()", FI_LI);
   if (VMSnok (status = sys$setprv (0, &AltpriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   /* anywhere near before starting logging! */
   NetGetServerHostName ();

   /* read the configuration file */
   Configure (&Config);

   if (Config.Busy) MaxConcurrentConnections = Config.Busy;

   if (CliServerPort)
      ServerPort = CliServerPort;
   else
   if (Config.ServerPort)
      ServerPort = Config.ServerPort;
   if (ServerPort < 1 || ServerPort > 65535)
      ErrorExitVmsStatus (NO_EXIT_MSG, "IP port", FI_LI);
   sprintf (ServerPortString, "%d", ServerPort);

   if (!CliLoggingDisabled && (Config.LoggingEnabled || CliLoggingEnabled))
      LoggingEnabled = true;
   else
      LoggingEnabled = false;

   if (!CliMonitorDisabled && (Config.MonitorEnabled || CliMonitorEnabled))
      MonitorEnabled = true;
   else
      MonitorEnabled = false;

   if (CliAcceptHostsPtr != NULL)
      Config.AcceptHostsPtr = CliAcceptHostsPtr;

   if (CliRejectHostsPtr != NULL)
      Config.RejectHostsPtr = CliRejectHostsPtr;

   /* get and set the service names and ports */
   NetGetService ();

   /* set the process name */
   if (VMSnok (status =
       sys$fao (&ProcessNameFaoDsc, &Length, &ProcessNameDsc, ServerPort)))
      ErrorExitVmsStatus (status, "sys$fao()", FI_LI);
   ProcessName[ProcessNameDsc.dsc$w_length = Length] = '\0';
   if (Debug) fprintf (stdout, "ProcessName |%s|\n", ProcessName);
   if (VMSnok (status = sys$setprn (&ProcessNameDsc)))
      ErrorExitVmsStatus (status, "sys$setprn()", FI_LI);

   /* ensure the GMT logical name is defined ok */
   if (VMSnok (status = TimeSetGmt ()))
      ErrorExitVmsStatus (status, "TimeSetGmt()", FI_LI);

   /* initialize message database */
   MsgLoad (&MsgDb);

   /* authentication/authorization configuration */
   AuthPathConfig (NULL);

   /* load the rule mapping database */
   MapUrl_Map (NULL, NULL, NULL, NULL, NULL);

   /* initialize DCL processing */
   DclInit ();

   /* initialize activity statistics */
   GraphActivityInit ();

   /* initialize file cache  */
   CacheInit (true);

   /* initialize the language environment */
   if (VMSok (status = lib$get_users_language (&NaturalLanguageDsc)))
   {
      for (cptr = NaturalLanguage; *cptr && !ISLWS(*cptr); cptr++);
      *cptr = '\0';
      if (strsame (NaturalLanguage, "ENGLISH"))
         NaturalLanguageEnglish = true;
      else
      {
         NaturalLanguageEnglish = false;
         fprintf (stdout, "%%%s-I-LANGUAGE, natural language is %s\n",
                  Utility, NaturalLanguage);
      }
   }
   else
   {
      if (status == LIB$_ENGLUSED)
      {
         NaturalLanguageEnglish = true;
         strcpy (NaturalLanguage, "ENGLISH");
         fprintf (stdout,
            "%%%s-W-LANGUAGE, natural language has defaulted to %s\n",
            Utility, NaturalLanguage);
      }
      else
         ErrorExitVmsStatus (status, "lib$get_users_language()", FI_LI);
   }

   /* reset logical names provided to the HTTPd monitor utility */
   if (MonitorEnabled)
      if (VMSnok (status = DefineMonitorLogicals (NULL)))
         ErrorExitVmsStatus (status, "DefineMonitorLogicals()", FI_LI);

   /* must be done after DefineMonitorLogicals, which reads previous values */
   if (AccountingZeroOnStartup || !MonitorEnabled)
   {
      /* initialize accounting information without reseting the logicals */
      ZeroAccounting (true);
   }

   /* set up to allow the HTTPD process to be controlled */
   if (VMSnok (status = ControlHttpd ()))
      ErrorExitVmsStatus (status, "ControlHttpd()", FI_LI);

   /* initialize the logging */
   if (LoggingEnabled)
      if (VMSnok (status = Logging (NULL, LOGGING_BEGIN)))
         ErrorExitVmsStatus (status, "Logging()", FI_LI);

   /* intialize request history mechanism (ensure it's reasonable!) */
   RequestHistoryMax = Config.RequestHistory;
   if (RequestHistoryMax > 999) RequestHistoryMax = 0;
   RequestHistoryCount = 0;
   RequestHistoryList.HeadPtr = RequestHistoryList.TailPtr = NULL;

   /* disable process swapping */
   if (NoSwapOut)
   {
      if (VMSnok (status = sys$setprv (1, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
      if (VMSnok (status = sys$setswm (1)))
         ErrorExitVmsStatus (status, "sys$setswm()", FI_LI);
      if (VMSnok (status = sys$setprv (0, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   }

   /* set up and declare the exit handler */
   ExitHandler.HandlerAddress = &HttpdExit;
   ExitHandler.ArgCount = 1;
   ExitHandler.ExitStatusPtr = &ExitStatus;
   if (VMSnok (status = sys$dclexh (&ExitHandler)))
      ErrorExitVmsStatus (status, "sys$dclexh()", FI_LI);

   /* effectively an infinite loop */
   exit (NetCreateService ());
}

/*****************************************************************************/
/*
Exit the HTTP server, via this declared exit handler.
*/

HttpdExit ()

{
   if (Debug) fprintf (stdout, "HttpdExit() %%X%08.08X\n", ExitStatus);

   /* run down any subprocesses associated with script processing */
   DclExit ();

   if (LoggingEnabled)
   {
      /* provide a server exit entry */
      Logging (NULL, LOGGING_END);
   }

   if (MonitorEnabled)
   {
      /* set exit status in logical names used by HTTPd monitor utility */
      Accounting.LastExitStatus = ExitStatus;
      DefineMonitorLogicals (NULL);
   }
}

/****************************************************************************/
/*
The 'rqptr->LocationPtr' is non-NULL indicating redirection.  This 
can be local, indicated by a leading '/', or it can be non-local, indicated
by anything other than a leading '/', usually a full URL including scheme
(e.g. "http:").

If authorization is enabled then all redirected requests should be returned
to the client in case authorization information needs to be applied to the
new request.  This may involve reformating a local redirection into a
non-local one.  If authentication is not enabled then check both local and
non-local redirections to see if it can be handled locally.

Return normal status to indicate local-redirection (and that the thread
should NOT be disposed of), or an error status to indicate client-handled,
non-local-redirection, or a legitimate error (and that the thread can go).
*/ 

int HttpdRedirect (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (BufferDsc, "");

   static char  ContentLength [32];
   static $DESCRIPTOR (contentLengthDsc, ContentLength);
   static $DESCRIPTOR (ContentLengthFaoDsc, "Content-Length: !UL\r\n\0");

   static $DESCRIPTOR (Http302FaoDsc,
"!AZ 302 Redirection.\r\n\
Location: !AZ\r\n\
Server: !AZ\r\n\
Date: !AZ\r\n\
\r\n\0");

   static $DESCRIPTOR (LocalFaoDsc,
"!AZ !AZ !AZ\r\n\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ\
!AZ\
\r\n\0");

   register unsigned long  *vecptr;
   register char  *cptr, *sptr;

   boolean  LocalRedirection;
   int  status,
        PortNumber;
   unsigned short  Length;
   unsigned long  FaoVector [64];
   char  *HttpMethodNamePtr;
   char  HttpMethodName [32],
         RedirectionUrl [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdRedirect() |%s|\n", rqptr->LocationPtr);

   if (rqptr->RedirectionCount++ > 1)
   {
      rqptr->LocationPtr = NULL;
      rqptr->ResponseStatusCode = 500;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_REDIRECTION), FI_LI);
      return (STS$K_ERROR);
   }

   rqptr->ResponseStatusCode = 302;

   /*
      A leading space is normally not possible.
      This indicates a specific HTTP method must be used.
      Format is: "<space>METHOD<space><redirection-path>".
   */
   if (rqptr->LocationPtr[0] == ' ')
   {
      HttpMethodNamePtr = sptr = HttpMethodName;
      for (cptr = rqptr->LocationPtr + 1;
           *cptr && *cptr != ' ';
           *sptr++ = *cptr++);
      *sptr = '\0';
      rqptr->LocationPtr = cptr + 1;
   }
   else
      HttpMethodNamePtr = rqptr->HttpMethodName;

   MapUrl_VirtualPath (rqptr->PathInfoPtr, rqptr->LocationPtr,
                       RedirectionUrl, sizeof(RedirectionUrl));

   /* now ignore old value of location redirection pointer */
   rqptr->LocationPtr = NULL;

#ifdef ALWAYS_LOCAL_REDIRECTION_IF_DETECTED

   LocalRedirection = false;

   if (RedirectionUrl[0] == '/')
      LocalRedirection = true;
   else
   {
      /*****************************/
      /* check for local host/port */
      /*****************************/

      if (!memcmp (cptr = RedirectionUrl, "http:", 5))
      {
         /* scan over the "http://" */
         while (*cptr && *cptr != '/') cptr++;
         while (*cptr && *cptr == '/') cptr++;
         if (Debug)
            fprintf (stdout, "cptr |%s|%s|\n", cptr, rqptr->ServerHostNamePtr);
         if (strsame (cptr,
                      rqptr->ServerHostNamePtr,
                      rqptr->ServerHostNameLength) &&
             (cptr[rqptr->ServerHostNameLength] == '/' ||
              cptr[rqptr->ServerHostNameLength] == ':'))
         {
            cptr += rqptr->ServerHostNameLength;
            if (*cptr == '/')
               LocalRedirection = true;
            else
            if (*cptr == ':')
            {
               PortNumber = atoi(++cptr);
               if (Debug) fprintf (stdout, "PortNumber: %d\n", PortNumber);
               if (PortNumber == ServerPort) LocalRedirection = true;
            }
            while (*cptr && *cptr != '/') cptr++;
         }

         if (LocalRedirection)
         {
            /* eliminate host details, copy just the path over the host detail */
            sptr = RedirectionUrl;
            while (*cptr) *sptr++ = *cptr++;
            *sptr = '\0';
            if (Debug) fprintf (stdout, "RedirectUrl |%s|\n", RedirectionUrl);
         }
      }
   }

#else

   if (RedirectionUrl[0] == '/')
      LocalRedirection = true;
   else
      LocalRedirection = false;

#endif

   if (LocalRedirection)
   {
      /*********************/
      /* local redirection */
      /*********************/

      Accounting.RedirectLocalCount++;

      vecptr = &FaoVector;

      *vecptr++ = HttpMethodNamePtr;
      *vecptr++ = RedirectionUrl;
      *vecptr++ = HttpProtocol;

      if (rqptr->HttpAcceptPtr != NULL)
      {
         *vecptr++ = "Accept: ";
         *vecptr++ = rqptr->HttpAcceptPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpAcceptLangPtr != NULL)
      {
         *vecptr++ = "Accept-Language: ";
         *vecptr++ = rqptr->HttpAcceptLangPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpAcceptCharsetPtr != NULL)
      {
         *vecptr++ = "Accept-Charset: ";
         *vecptr++ = rqptr->HttpAcceptCharsetPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpUserAgentPtr != NULL)
      {
         *vecptr++ = "User-Agent: ";
         *vecptr++ = rqptr->HttpUserAgentPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpRefererPtr != NULL)
      {
         *vecptr++ = "Referer: ";
         *vecptr++ = rqptr->HttpRefererPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpHostPtr != NULL)
      {
         *vecptr++ = "Host: ";
         *vecptr++ = rqptr->HttpHostPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpAuthorizationPtr != NULL)
      {
         *vecptr++ = "Authorization: ";
         *vecptr++ = rqptr->HttpAuthorizationPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpIfModifiedSincePtr != NULL)
      {
         *vecptr++ = "If-Modified-Since: ";
         *vecptr++ = rqptr->HttpIfModifiedSincePtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpPragmaPtr)
      {
         *vecptr++ = "Pragma: ";
         *vecptr++ = rqptr->HttpPragmaPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpCookiePtr)
      {
         *vecptr++ = "Cookie: ";
         *vecptr++ = rqptr->HttpCookiePtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->ContentLength)
      {
         sys$fao (&ContentLengthFaoDsc, 0, ContentLength,
                  rqptr->ContentLength);
         *vecptr++ = ContentLength;
      }
      else
         *vecptr++ = "";

      if (rqptr->KeepAliveRequest)
         *vecptr++ = "Connection: Keep-Alive\r\n";
      else
         *vecptr++ = "";

      if (rqptr->HttpForwardedPtr != NULL)
      {
         *vecptr++ = "Forwarded: ";
         *vecptr++ = rqptr->HttpForwardedPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      BufferDsc.dsc$a_pointer = rqptr->NetReadBufferPtr;
      BufferDsc.dsc$w_length = rqptr->NetReadBufferSize;

      status = sys$faol (&LocalFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->LocationPtr = NULL;
         rqptr->ErrorTextPtr = "sys$faol()";
         rqptr->ResponseStatusCode = 500;
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (STS$K_ERROR);
      }

      if (Debug) fprintf (stdout, "|%s|\n", rqptr->NetReadBufferPtr);
      rqptr->RequestHeaderPtr = rqptr->NetReadBufferPtr;
      rqptr->RequestHeaderLength = Length;

      /* normal status indicates its a local redirection */
      return (SS$_NORMAL);
   }
   else
   {
      /*************************/
      /* non-local redirection */
      /*************************/

      Accounting.RedirectRemoteCount++;

      BufferDsc.dsc$a_pointer = rqptr->NetReadBufferPtr;
      BufferDsc.dsc$w_length = rqptr->NetReadBufferSize;

      status = sys$fao (&Http302FaoDsc, &Length, &BufferDsc,
                        HttpProtocol, RedirectionUrl, SoftwareID,
                        rqptr->GmDateTime);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->LocationPtr = NULL;
         rqptr->ErrorTextPtr = "sys$fao()";
         rqptr->ResponseStatusCode = 500;
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (STS$K_ERROR);
      }
      if (Debug) fprintf (stdout, "|%s|\n", rqptr->NetReadBufferPtr);
      NetWrite (rqptr, &ConcludeProcessing, rqptr->NetReadBufferPtr, Length-1);

      /* unreachable indicates it's non-local redirection :^) */
      return (SS$_UNREACHABLE);
   }
}

/*****************************************************************************/
/*
This function can be called from two sources.  First, NetAccept() where a new
connection has been established and memory has just been allocated for this
connection's thread structure.  Second, from ConcludeProcessing() where a
persistent connection (HTTP/1.0) is involved.

Both then queue a read from the network connection which calls an AST
function when the read completes, to process the request.  The keep-alive
read usually has a far shorter timeout associated with it so that rapid,
sequential reads (as happens when a document is being populated with images
for instance) are serviced from the one connection, but these are not left
open consuming resources on the off-chance of another request in the
not-too-distance future.

Persistent connections are implemented by disposing of all memory allocated
on a request's heap, and all data in the connection thread structure by
zeroing, EXCEPT that above the field 'RetainAboveZeroBelow' (client
connection information), then treating the almost-amnesic structure like a
brand-new connection (only with a shorter timeout period).
*/ 

HttpdConnection
(
struct RequestStruct *rqptr,
boolean  NewConnection
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdConnection()\n");

   if (NewConnection)
   {
      /******************/
      /* new connection */
      /******************/

      /* add entry to the top of the request list */
      ListAddHead (&RequestList, &rqptr->RequestListEntry, rqptr);

      /* timestamp the transaction */
      sys$gettim (&rqptr->BinaryTime);
      sys$numtim (&rqptr->NumericTime, &rqptr->BinaryTime);
      HttpGmTimeString (rqptr->GmDateTime, &rqptr->BinaryTime);

      /* initialize the timer for input */
      HttpdTimerSet (rqptr, TIMER_INPUT);

      /* initalize the HTTPd supervisor (if it's not already running) */
      HttpdSupervisor (false);
   }
   else
   {
      /*************************/
      /* persistent connection */
      /*************************/

      memset ((unsigned char*)&rqptr->RetainAboveZeroBelow,
              0,
              ((unsigned char*)&rqptr->ZeroedEnd -
               (unsigned char*)&rqptr->RetainAboveZeroBelow));

      /*
         Timestamp the keep-alive period for AdminRequestReport()
         Transaction will again be time-stamped if/when request received!
      */
      sys$gettim (&rqptr->BinaryTime);

      /* initialize the timer for keep-alive */
      HttpdTimerSet (rqptr, TIMER_KEEPALIVE);
   }

   if (Debug) fprintf (stdout, "KeepAliveCount: %d\n", rqptr->KeepAliveCount);

   /* 
      Queue an asynchronous read to get the request from the client.
      When the read completes call RequestGet() completion function to
      further process the request.  Allow one byte for a terminating null.
   */
   if ((rqptr->NetReadBufferPtr =
       HeapAlloc (rqptr, NetReadBufferSize+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, FI_LI);
      return (STS$K_ERROR);
   }
   rqptr->NetReadBufferSize = NetReadBufferSize;

   NetRead (rqptr, &RequestGet,
            rqptr->NetReadBufferPtr, rqptr->NetReadBufferSize);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function may be AST-delivery executed ONE OR MORE TIMES, FROM ITSELF, 
before a connection is actually disposed of.

If a next-task function was specified, and there has been no error message
generated, then execute it.

If there are any characters remaining in the output buffer then flush it with
an AST leading straight back to this function!

If an error has been generated then send it to the client once again with an
AST leading straight back to this function!

Check for redirection.  If local redirection do NOT dispose of the thread
(we'll be re-using the thread context).

If a persistent connection then treat it like a new connection and go back
to listening for the next request.

Otherwise close the client network socket, release dynamic data structures
allocated, etc.
*/

ConcludeProcessing (struct RequestStruct *rqptr)

{
   static long  Addx2 = 2,
                Pid = 0,
                Subx2 = 2,
                EdivTen = 10;
   static unsigned long  JpiCurPriv [2];
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiCurPriv), JPI$_CURPRIV, &JpiCurPriv, 0 },
      {0,0,0,0}
   };

   int  status,
        StatusCodeGroup;
   unsigned long  Remainder;
   unsigned long  BinaryTime [2],
                  QuadScratch [2],
                  ResultTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConcludeProcessing()\n");

   /*
      Flush anything currently remaining in the thread's output buffer.
      AST back to this same function to continue disposing of the thread!
   */
   if (rqptr->OutputBufferCurrentPtr > rqptr->OutputBufferPtr)
   {
      NetWriteBuffered (rqptr, &ConcludeProcessing, NULL, 0);
      return;
   }

   /*
      If an error has been reported send this to the client.
      AST back to this same function to continue disposing of the thread!
   */
   if (rqptr->ErrorMessagePtr != NULL)
      if (VMSok (status = ErrorSendToClient (rqptr, &ConcludeProcessing)))
         return;

   /*
      If the response header has not yet been sent then do it.
      AST back to this same function to continue disposing of the thread!
   */
   if (!rqptr->BytesTx && rqptr->ResponseHeaderPtr != NULL)
   {
      NetWrite (rqptr, &ConcludeProcessing, 0, 0);
      return;
   }

   /*************************/
   /* response elapsed time */
   /*************************/

   if (!rqptr->KeepAliveTimeout)
   {
      sys$gettim (&BinaryTime);
      status = lib$subx (&BinaryTime, &rqptr->BinaryTime, 
                         &ResultTime, &Subx2);
      /* convert to microseconds (longword becomes 71 minutes max) */
      if (VMSok (status))
         status = lib$ediv (&EdivTen, &ResultTime,
                            &rqptr->ResponseDuration, &Remainder);
      if (Debug) fprintf (stdout, "elapsed: %%X%08.08X\n", status);
      if (VMSnok (status)) rqptr->ResponseDuration = 0;

      Accounting.ResponseDurationCount++;
      QuadScratch[0] = rqptr->ResponseDuration;
      QuadScratch[1] = 0;
      lib$addx (&QuadScratch, &Accounting.QuadResponseDuration,
                &Accounting.QuadResponseDuration, &Addx2);
      if (rqptr->ResponseDuration < Accounting.ResponseDurationMin)
         Accounting.ResponseDurationMin = rqptr->ResponseDuration;
      if (rqptr->ResponseDuration > Accounting.ResponseDurationMax)
         Accounting.ResponseDurationMax = rqptr->ResponseDuration;
   }

   /***************/
   /* redirection */
   /***************/

   if (rqptr->LocationPtr != NULL)
   {
      /* if local do NOT dispose of thread */
      if (VMSok (status = HttpdRedirect (rqptr)))
      {
         /* log the original request, before generating it redirected */
         if (LoggingEnabled) Logging (rqptr, LOGGING_ENTRY);

         /* re-call the parsing function to begin re-processing the request */
         RequestParseAndExecute (rqptr);
         return;
      }

      /*
         If a genuine error reported when redirecting send this to the client.
         AST back to this same function to continue disposing of the thread!
      */
      if (rqptr->ErrorMessagePtr != NULL)
      {
         if (VMSok (ErrorSendToClient (rqptr, &ConcludeProcessing)))
            return;
      }
      else
      {
         /*
            A status of SS$_UNREACHABLE is returned when non-local redirect.
            The network write will AST back to this same function to continue.
         */
         if (status == SS$_UNREACHABLE) return;
      }
   }

   /**************************/
   /* dispose of the REQUEST */
   /**************************/

   if (Debug) fprintf (stdout, "rqptr->BytesTx: %d\n", rqptr->BytesTx);

   /* a little more accounting */
   Accounting.ConnectCurrent = CurrentConnectCount-1;
   QuadScratch[0] = rqptr->BytesTx;
   QuadScratch[1] = 0;
   lib$addx (&QuadScratch, &Accounting.QuadBytesTx,
             &Accounting.QuadBytesTx, &Addx2);
   QuadScratch[0] = rqptr->BytesRx;
   QuadScratch[1] = 0;
   lib$addx (&QuadScratch, &Accounting.QuadBytesRx,
             &Accounting.QuadBytesRx, &Addx2);

   /* set logical names used by HTTPd monitor utility */
   if (MonitorEnabled) DefineMonitorLogicals (rqptr);

   if (!rqptr->KeepAliveTimeout)
   {
      StatusCodeGroup = rqptr->ResponseStatusCode / 100;
      if (StatusCodeGroup < 0 || StatusCodeGroup > 5) StatusCodeGroup = 0;
      Accounting.ResponseStatusCodeCountArray[StatusCodeGroup]++;

      if (LoggingEnabled) Logging (rqptr, LOGGING_ENTRY);

      /* if a request history is being kept then provide the entry */
      if (RequestHistoryMax) HttpdRequestHistory (rqptr);

      /* if keeping activity statistics */
      if (Config.ActivityNumberOfDays) GraphActivityIncrement (rqptr);
   }

   /* free dynamic memory allocated for the request's heap */
   HeapFree (rqptr);

   if (rqptr->KeepAliveRequest &&
       rqptr->KeepAliveResponse &&
       rqptr->KeepAliveCount < MaxKeepAliveCount)
   {
      /*************************/
      /* PERSISTENT connection */
      /*************************/

      /*
         If the persistent connection is accepted then just return.
         Otherwise continue to completely dispose of the connection.
      */
      if (VMSok (HttpdConnection (rqptr, false))) return;

      /* free any dynamic memory that may have been reallocated to the heap */
      HeapFree (rqptr);
   }

   /*****************************/
   /* dispose of the CONNECTION */
   /*****************************/

   /* remove from the request list */
   ListRemove (&RequestList, &rqptr->RequestListEntry);

#ifdef IP_UCX
   NetShutdownSocket (rqptr->ClientChannel);
#endif

#ifdef IP_NETLIB
   NetShutdownSocket (&rqptr->ClientNetLibSocket);
#endif

   /* free dynamic memory allocated for the connection structure */
   if (Debug) fprintf (stdout, "free() rqptr\n");
   free (rqptr);

   /* one less to worry about! */
   if (CurrentConnectCount) CurrentConnectCount--;

   /* don't bother checking for exit or restart if connections still exist */
   if (CurrentConnectCount) return;

   /* if the control module has requested server exit or restart */
   if (ControlExitRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server exit\n", Utility);
      ExitStatus = SS$_NORMAL;
      HttpdExit();
      sys$delprc (0, 0);
   }
   else
   if (ControlRestartRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server restart\n", Utility);
      exit (SS$_NORMAL);
   }

   /*
       As the code gets more complex it's becoming increasingly possible
       a coding or design error will leave privileges turned on somewhere.
       To help detect any such problem ensure everthing's off occasionlly
       (when there are no more connections to process).
   */
   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpi()", FI_LI);

   if ((JpiCurPriv[0] & ~AvJoePrivMask[0]) || JpiCurPriv[1])
   {
      if (Debug)
         fprintf (stdout, "%08.08X%08.08X\n", JpiCurPriv[1], JpiCurPriv[0]);
      ErrorExitVmsStatus (NO_EXIT_MSG, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Keeps a linked list history from most to least recent that can be reported on
by AdminRequestReport().
*/

int HttpdRequestHistory (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr, *zptr;
   register struct RequestHistoryStruct  *hlptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdRequestHistory()\n");

   if (RequestHistoryCount < RequestHistoryMax)
   {
      /* allocate memory for a new entry */
      if ((hlptr = calloc (1, sizeof (struct RequestHistoryStruct))) == NULL)
         ErrorExitVmsStatus (status, "calloc()", FI_LI);
      RequestHistoryCount++;
   }
   else
   {
      /* reuse the tail entry (least recent) */
      hlptr = RequestHistoryList.TailPtr;
      ListRemove (&RequestHistoryList, &hlptr->HistoryListEntry);
   }

   /* add entry to the head of the history list (most recent) */
   ListAddHead (&RequestHistoryList, &hlptr->HistoryListEntry, hlptr);

   memcpy (&hlptr->BinaryTime, &rqptr->BinaryTime, 8);
   hlptr->BytesRx = rqptr->BytesRx;
   hlptr->BytesTx = rqptr->BytesTx;
   hlptr->ResponseDuration = rqptr->ResponseDuration;
   hlptr->ResponseStatusCode = rqptr->ResponseStatusCode;

   zptr = (sptr = hlptr->ClientAndRequest) + sizeof(hlptr->ClientAndRequest);

   if (rqptr->RemoteUser[0])
   {
      cptr = rqptr->RemoteUser;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = '@';
   }
   cptr = rqptr->ClientHostName;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr < zptr) *sptr++ = '\0';
   hlptr->RequestPtr = sptr;

   if (rqptr->PathInfoPtr != NULL && rqptr->QueryStringPtr != NULL)
   {
      cptr = rqptr->HttpMethodName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = ' ';
      cptr = rqptr->ScriptName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr = rqptr->PathInfoPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (rqptr->QueryStringPtr[0] && sptr < zptr) *sptr++ = '?';
      cptr = rqptr->QueryStringPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }

   if (sptr >= zptr)
   {
      hlptr->Truncated = true;
      sptr--;
   }
   else
      hlptr->Truncated = false;
   *sptr = '\0';
}

/*****************************************************************************/
/*
Set the timer counters for the request according to the function code. 
HttpdSupervisor() will monitor these counters.  Due to the way the seconds are
counted-down, and the setting of a timer may be in the middle of any one
second interval, increment very small periods to ensure they are at least that.
*/

int HttpdTimerSet
(
struct RequestStruct *rqptr,
int Function
)
{
   static boolean  Initialize = true;
   static unsigned long  InputSeconds,
                         OutputSeconds,
                         KeepAliveSeconds;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdTimerSet() %d\n", Function);

   if (Initialize)
   {
      Initialize = false;
      if (Config.InputTimeoutMinutes)
         InputSeconds = Config.InputTimeoutMinutes * 60;
      else
         InputSeconds = DefaultInputTimeoutMinutes * 60;
      if (Config.OutputTimeoutMinutes)
         OutputSeconds = Config.OutputTimeoutMinutes * 60;
      else
         OutputSeconds = DefaultOutputTimeoutMinutes * 60;
      if (Config.KeepAliveTimeoutSeconds)
         KeepAliveSeconds = Config.KeepAliveTimeoutSeconds;
      else
         KeepAliveSeconds = DefaultKeepAliveTimeoutSeconds;
      /* ensure small periods are at least that duration */
      if (KeepAliveSeconds && KeepAliveSeconds < 10) KeepAliveSeconds++;
   }

   switch (Function)
   {
      case TIMER_INPUT :
         rqptr->TimerOutputCount = rqptr->TimerKeepAliveCount = 0;
         rqptr->TimerInputCount = InputSeconds;
         return;

      case TIMER_OUTPUT :
         rqptr->TimerInputCount = rqptr->TimerKeepAliveCount = 0;
         rqptr->TimerOutputCount = OutputSeconds;
         return;

      case TIMER_KEEPALIVE :
         rqptr->TimerInputCount = rqptr->TimerOutputCount = 0;
         rqptr->TimerKeepAliveCount = KeepAliveSeconds;
         return;

      default :
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Whenever at least one request is being processed this function is called every
second to supervise the requests' progress.  Scanning the list of current
processes it decrements set timer counters.  When a counter reaches zero
network I/O or DCL processing is cancelled.  This results in the request being
terminated.

This function was introduced to address a suspected multi-thread timing problem
associated with setting and cancelling independent timers and ASTs for each
request.  It is probably no more expensive than setting and reseting VMS
timers for each request's processing, and does seem to give some measure (or
feeling of) more structure and control!
*/

HttpdSupervisor (boolean TimerExpired)

{
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };

#ifdef IP_NETLIB
   static unsigned long  NetLibShutdown = NETLIB_K_SHUTDOWN_BOTH;
#endif

   static boolean  TimerSet = false;

   register struct RequestStruct  *rqeptr;
   register struct GenericListEntryStruct  *leptr;

   int  status;
   struct AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdSupervisor() %d\n", TimerExpired);

   if (Debug)
      fprintf (stdout, "Head %d\nTail %d\n",
               RequestList.HeadPtr, RequestList.TailPtr);

   if (!TimerExpired)
   {
      /* new request, if the supervisor is already running just return */
      if (TimerSet) return;

      /* kick the supervisor timer into life, and then return */
      if (VMSnok (status =
          sys$setimr (0, &OneSecondDelta, &HttpdSupervisor, true, 0)))
         ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
      TimerSet = true;
      return;
   }

   /* must have got here via the supervisor's timer expiring */
   TimerSet = false;

   /* if nothing in the current request list return without sys$setimr() */
   if ((leptr = RequestList.HeadPtr) == NULL) return;

   /* process the current request list entries */
   for (/* from above */; leptr != NULL; leptr = leptr->NextPtr)
   {
      rqeptr = (struct RequestStruct*)leptr->DataPtr;

      if (Debug)
         fprintf (stdout, "%d <- %d -> %d (%d) in: %d out: %d alive: %d\n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqeptr,
                  rqeptr->TimerInputCount, rqeptr->TimerOutputCount,
                  rqeptr->TimerKeepAliveCount);

      /* ordered in the most common occurance of timings */
      if (rqeptr->TimerOutputCount)
         if (--rqeptr->TimerOutputCount)
            continue;
         else;
      else
      if (rqeptr->TimerKeepAliveCount)
         if (--rqeptr->TimerKeepAliveCount)
            continue;
         else;
      else
      if (rqeptr->TimerInputCount)
         if (--rqeptr->TimerInputCount)
            continue;

      if (Debug) fprintf (stdout, "terminate %d\n", rqeptr);

      if (rqeptr->DclTaskPtr != NULL)
      {
         /* DCL involved, do any DCL-specific cleanup */
         DclConcludeTask (rqeptr->DclTaskPtr, true);
      }
      else
      {
#ifdef IP_UCX
         /* cancel network I/O, should force the connection to abort */
         sys$cancel (rqeptr->ClientChannel);
#endif

#ifdef IP_NETLIB
         /* shutdown socket fully, should force the connection to abort */
         status = netlib_shutdown (&rqeptr->ClientNetLibSocket,
                                   &NetLibShutdown, &IOsb, 0, 0);
         if (Debug)
            fprintf (stdout, "netlib_shutdown %%X%08.08X IOsb %%X%08.08X\n",
                     status, IOsb.Status);
#endif
      }
   }

   if (VMSnok (status =
       sys$setimr (0, &OneSecondDelta, &HttpdSupervisor, true, 0)))
      ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);

   TimerSet = true;
}

/****************************************************************************/

